theory concur_history_relations
imports Main
begin

definition abcd :: "'a rel \<Rightarrow> bool"
where
  "abcd r \<longleftrightarrow> (\<forall>a b c d. (a, b) \<in> r \<longrightarrow> (c, d) : r \<longrightarrow> (a, d)\<in>r \<or> (c, b)\<in>r)"

lemma abcdI : "(\<And>a b c d. (a, b) \<in> r \<Longrightarrow> (c, d) \<in> r \<Longrightarrow> (a, d) \<in> r \<or> (c, b) \<in> r) \<Longrightarrow> abcd r"
by (metis abcd_def)


lemma abcdD : "abcd r \<Longrightarrow> (a, b) \<in> r \<Longrightarrow> (c, d) \<in> r \<Longrightarrow> (a, d) \<in> r \<or> (c, b) \<in> r"
by (metis abcd_def)

lemma lem_total_then_abcd : "total r \<Longrightarrow> trans r \<Longrightarrow> abcd r"
apply(unfold abcd_def)
by (metis UNIV_I total_on_def transD)

lemma lem_transitive_irreflexive_then_acyclic : "trans r \<Longrightarrow> irrefl r \<Longrightarrow> acyclic r"
by (metis acyclic_irrefl trancl_id)

lemma lem_irreflexive_then_not_equal : "irrefl r \<Longrightarrow> (a, b) \<in> r \<Longrightarrow> a \<noteq> b"
by (metis irrefl_def)

(* A concurrent history is a set of operations with precedence order pre. The precedence order is a
   partial order with the abcd property.*)
locale conc_history =
  fixes pre :: "'a rel"
  assumes pre_transitive: "trans pre"
  assumes pre_antisymmetric: "antisym pre"
  assumes pre_irreflexive: "irrefl pre"
  assumes abcd : "abcd pre"

lemma conc_historyI : "trans pre \<Longrightarrow> antisym pre \<Longrightarrow> irrefl pre \<Longrightarrow> abcd pre \<Longrightarrow> conc_history pre"
apply (unfold conc_history_def)
by metis

lemma conc_historyD : "conc_history pre \<Longrightarrow> trans pre \<and> antisym pre \<and> irrefl pre \<and> abcd pre"
apply (unfold conc_history_def)
by metis

(* I also define the value relation with a predicate. *)
definition val_order :: "'a rel \<Rightarrow> bool"
where
  "val_order val \<longleftrightarrow> (\<forall> ia ra. (ia, ra) \<in> val \<longrightarrow> \<not>(\<exists> b. (b, ia) \<in> val \<or> (ra, b) \<in> val))"

(* The properties of a wellformed history are also defined using predicates.
   We talked about this in York and you suggested that I could also make one
   locale per property, but after I started defining the precedence order with
   a predicate I just continued with it. *)
definition push_once :: "'a rel \<Rightarrow> bool"
where
  "push_once val \<longleftrightarrow> (\<forall> ia ib ra. (ia, ra) \<in> val \<longrightarrow> (ib, ra) \<in> val \<longrightarrow> ia = ib)"

definition pop_once :: "'a rel \<Rightarrow> bool"
where
  "pop_once val \<longleftrightarrow> (\<forall> ia ra rb. (ia, ra) \<in> val \<longrightarrow> (ia, rb) \<in> val \<longrightarrow> ra = rb)"

definition val_then_not_pr :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where
  "val_then_not_pr pre val \<longleftrightarrow> (\<forall> ia ra. (ia, ra) \<in> val \<longrightarrow>(ra, ia)\<notin>pre)"

lemma val_then_not_prI : "\<forall> ia ra . (ia, ra) \<in> val \<longrightarrow> (ra, ia) \<notin> pre \<Longrightarrow> val_then_not_pr pre val"
by (metis val_then_not_pr_def)

definition isPush :: "'a rel \<Rightarrow> 'a \<Rightarrow> bool"
where
  "isPush val a \<longleftrightarrow> (\<exists> b . (a, b) \<in> val)"

lemma isPushD : "isPush val a \<Longrightarrow> \<exists> b . (a, b) \<in> val"
by (metis isPush_def)

lemma isPushI : "(a, b) \<in> val \<Longrightarrow> isPush val a"
by (metis isPush_def)

definition isPop :: "'a rel \<Rightarrow> 'a \<Rightarrow> bool"
where
  "isPop val a \<longleftrightarrow> (\<exists> b . (b, a) \<in> val)"

lemma isPopD : "isPop val a \<Longrightarrow> \<exists> b . (b, a) \<in> val"
by (metis isPop_def)

lemma isPopI : "(a, b) \<in> val \<Longrightarrow> isPop val b"
by (metis isPop_def)


definition sameVal :: "'a rel \<Rightarrow> 'a rel"
where "sameVal val = {(a, b) . (a, b) \<in> val \<or> (b, a) \<in> val}"

lemma sameValD : "(a, b) \<in> sameVal val \<Longrightarrow> (a, b) \<in> val \<or> (b, a) \<in> val"
apply (unfold sameVal_def)
by (simp add : set_eq_iff)

lemma sameValI1 : "(a, b) \<in> val \<Longrightarrow> (a, b) \<in> sameVal val"
apply (unfold sameVal_def)
by (simp add : set_eq_iff)

lemma sameValI2 : "(a, b) \<in> val \<Longrightarrow> (b, a) \<in> sameVal val"
apply (unfold sameVal_def)
by (simp add : set_eq_iff)

lemma sameVal_commutative : "(a, b) \<in> sameVal val \<Longrightarrow> (b, a) \<in> sameVal val"
by (metis sameValD sameValI1 sameValI2)

definition elimination_free :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a set"
where
  "elimination_free pre val = {a . (\<exists> b . (a, b) \<in> sameVal val \<and> ((a, b) \<in> pre \<or> (b, a) \<in> pre))}"

lemma elimination_freeD : "a \<in> elimination_free pre val  \<Longrightarrow> \<exists> b . (a, b) \<in> sameVal val \<and> ((a, b) \<in> pre \<or> (b, a) \<in> pre)"
apply (unfold elimination_free_def)
by (simp add: set_eq_iff)

lemma elimination_freeI1 : "(a, b) \<in> sameVal val \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> a \<in> elimination_free pre val"
apply (unfold elimination_free_def)
apply (simp add: set_eq_iff)
by metis

lemma elimination_free_in_push_pop : "a \<in> elimination_free pre val \<Longrightarrow> a \<in> {a . isPush val a \<or> isPop val a}"
apply (simp add: set_eq_iff)
by (metis elimination_freeD isPop_def isPush_def sameValD)

lemma elimination_free_subset_push_pop : "elimination_free pre val \<subseteq> {a . isPush val a \<or> isPop val a}"
by (metis elimination_free_in_push_pop subsetI)

lemma finite_elimination_free : "finite {a . isPush val a \<or> isPop val a} \<Longrightarrow> finite (elimination_free pre val)"
by (metis elimination_free_subset_push_pop finite_subset)

lemma elimination_freeI2 : "(a, b) \<in> sameVal val \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> b \<in> elimination_free pre val"
apply (unfold elimination_free_def)
apply (simp add: set_eq_iff)
by (metis sameValD sameValI1 sameValI2)

definition elimination :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a set"
  where "elimination pre val = {a . (isPush val a \<or> isPop val a) \<and> a \<notin> elimination_free pre val}"

lemma eliminationD : "a \<in> elimination pre val \<Longrightarrow> \<exists> b . (a, b) \<in> sameVal val \<and> (a, b) \<notin> pre \<and> (b, a) \<notin> pre"
apply (unfold elimination_def)
apply (simp add:set_eq_iff)
by (metis elimination_freeI1 elimination_freeI2 isPop_def isPush_def sameValI1 sameValI2)

definition elimination_free_pre :: "'a rel \<Rightarrow> 'a rel \<Rightarrow>'a rel"
where
  "elimination_free_pre pre val = {(a, b) . a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val \<and> (a, b) \<in> pre}"

lemma elimination_free_preD : "(a, b) \<in> elimination_free_pre pre val \<Longrightarrow> a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val \<and> (a, b) \<in> pre"
apply (unfold elimination_free_pre_def)
by (simp add: set_eq_iff)

lemma elimination_free_preI : "a \<in> elimination_free pre val \<Longrightarrow> b \<in> elimination_free pre val \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> elimination_free_pre pre val"
apply (unfold elimination_free_pre_def)
by (simp add: set_eq_iff)

definition elimination_free_val :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel"
where
  "elimination_free_val pre val = {(a, b) . a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val \<and> (a, b) \<in> val}"

lemma elimination_free_valD : "(a, b) \<in> elimination_free_val pre val \<Longrightarrow> a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val \<and> (a, b) \<in> val"
apply (unfold elimination_free_val_def)
by (simp add: set_eq_iff)

lemma elimination_free_valI : "a \<in> elimination_free pre val \<Longrightarrow> b \<in> elimination_free pre val \<Longrightarrow> (a, b) \<in> val \<Longrightarrow> (a, b) \<in> elimination_free_val pre val"
apply (unfold elimination_free_val_def)
by (simp add: set_eq_iff)

definition elimination_free_ir :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel"
where
  "elimination_free_ir pre val ir = {(a, b) . a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val \<and> (a, b) \<in> ir}"

lemma elimination_free_irD : "(a, b) \<in> elimination_free_ir pre val ir \<Longrightarrow> a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val \<and> (a, b) \<in> ir"
apply (unfold elimination_free_ir_def)
by (simp add : set_eq_iff)

lemma elimination_free_irI : "a \<in> elimination_free pre val \<Longrightarrow> b \<in> elimination_free pre val \<Longrightarrow> (a, b) \<in> ir \<Longrightarrow> (a, b) \<in> elimination_free_ir pre val ir"
apply (unfold elimination_free_ir_def)
by (simp add: set_eq_iff)

definition ins_order :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel"
where
  "(ins_order pre val ir) = {(a, b) . isPush val a \<and> isPush val b \<and> ((a, b) \<in> pre \<or> (\<exists> c . (a, c) \<in> pre \<and> (c, b) \<in> ir))}"

lemma ins_orderI1 : "isPush val a \<Longrightarrow> isPush val b \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> ins_order pre val ir"
apply (unfold ins_order_def)
by (simp add: set_eq_iff)

lemma ins_orderI2 : "isPush val a \<Longrightarrow> isPush val b \<Longrightarrow> (a, c) \<in> pre \<Longrightarrow> (c, b) \<in> ir \<Longrightarrow> (a, b) \<in> ins_order pre val ir"
apply (unfold ins_order_def)
apply (simp add: set_eq_iff)
by metis

lemma ins_orderD : "(a, b) \<in> (ins_order pre val ir) \<Longrightarrow> isPush val a \<and> isPush val b \<and> ((a, b) \<in> pre \<or>(\<exists> c . (a, c) \<in> pre \<and> (c, b) \<in> ir))"
apply (unfold ins_order_def)
by (simp add: set_eq_iff)

definition rem_order :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel"
where
  "(rem_order pre val ir) = {(a, b) . isPop val a \<and> isPop val b \<and> ((a, b) \<in> pre \<or> (\<exists> c . (a, c) \<in> ir \<and> (c, b) \<in> ir))}"

lemma rem_orderI1 : "isPop val a \<Longrightarrow> isPop val b \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> rem_order pre val ir"
apply (unfold rem_order_def)
by (simp add: set_eq_iff)

lemma rem_orderI2 : "isPop val a \<Longrightarrow> isPop val b \<Longrightarrow> (a, c) \<in> ir \<Longrightarrow> (c, b) \<in> ir \<Longrightarrow> (a, b) \<in> rem_order pre val ir"
apply (unfold rem_order_def)
apply (simp add: set_eq_iff)
by metis

lemma rem_orderD : "(a, b) \<in> (rem_order pre val ir) \<Longrightarrow> isPop val a \<and> isPop val b \<and> ((a, b) \<in> pre \<or>(\<exists> c . (a, c) \<in> ir \<and> (c, b) \<in> ir))"
apply (unfold rem_order_def)
by (simp add: set_eq_iff)

definition alternating :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where
  " alternating val ir \<longleftrightarrow> (\<forall> a b . (a, b) \<in> ir \<longrightarrow> (isPush val a \<and> isPop val b) \<or> (isPop val a \<and> isPush val b)) 
                                   \<and> (\<forall> a b . isPush val a \<longrightarrow> isPop val b \<longrightarrow> (a, b) \<in> ir \<or> (b, a) \<in> ir)"

lemma alternatingD : "alternating val ir \<Longrightarrow> (\<forall> a b . (a, b) \<in> ir \<longrightarrow> (isPush val a \<and> isPop val b) \<or> (isPop val a \<and> isPush val b)) 
                                   \<and> (\<forall> a b . isPush val a \<longrightarrow> isPop val b \<longrightarrow> (a, b) \<in> ir \<or> (b, a) \<in> ir)"
by (smt alternating_def)

lemma alternatingI : "(\<forall> a b . (a, b) \<in> ir \<longrightarrow> (isPush val a \<and> isPop val b) \<or> (isPop val a \<and> isPush val b)) 
                                   \<Longrightarrow> (\<forall> a b . isPush val a \<longrightarrow> isPop val b \<longrightarrow> (a, b) \<in> ir \<or> (b, a) \<in> ir)
                                   \<Longrightarrow> alternating val ir"
by (metis (full_types) alternating_def)

definition order_correct :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where
  "order_correct pre val \<longleftrightarrow> (\<exists> ir . alternating val ir \<and> (acyclic (pre \<union> ir) \<and>
    (\<forall> ia ra ib . (ia, ra) \<in> val \<and> (ia, ra) \<in> pre  \<and> 
       (ia, ib) \<in> (ins_order pre val ir) \<and> (ib, ra) \<in> ir \<longrightarrow> 
            (\<exists> rb . (ib, rb) \<in> val \<and> (ra, rb) \<notin> (rem_order pre val ir)))))"

(* definition order_correct :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where
  "order_correct pre val \<longleftrightarrow> (\<exists> ir . alternating val ir \<and> (acyclic (pre \<union> ir) \<and>
    (\<forall> ia ra ib . (ia, ra) \<in> val \<and> (ia, ra) \<in> pre \<and> ib \<in> elimination_free pre val \<and> 
       (ia, ib) \<in> (ins_order pre val ir) \<and> (ib, ra) \<in> ir \<longrightarrow> 
            (\<exists> rb . (ib, rb) \<in> val \<and> (ra, rb) \<notin> (rem_order pre val ir)))))"
*)

lemma order_correctD : "order_correct pre val \<Longrightarrow> (\<exists> ir . alternating val ir \<and> (acyclic (pre \<union> ir) \<and>
    (\<forall> ia ra ib . (ia, ra) \<in> val \<and> (ia, ra) \<in> pre \<and> 
       (ia, ib) \<in> (ins_order pre val ir) \<and> (ib, ra) \<in> ir \<longrightarrow> 
            (\<exists> rb . (ib, rb) \<in> val \<and> (ra, rb) \<notin> (rem_order pre val ir)))))"
by (metis order_correct_def)

lemma order_correctI : "alternating val ir \<Longrightarrow> acyclic (pre \<union> ir) \<Longrightarrow> 
    (\<forall> ia ra ib . (ia, ra) \<in> val \<and> (ia, ra) \<in> pre \<and> 
       (ia, ib) \<in> (ins_order pre val ir) \<and> (ib, ra) \<in> ir \<longrightarrow> 
            (\<exists> rb . (ib, rb) \<in> val \<and> (ra, rb) \<notin> (rem_order pre val ir))) \<Longrightarrow> order_correct pre val"
apply (unfold order_correct_def)
apply (rule_tac x="ir" in exI)
by (metis (erased, hide_lams))

(* Definition of a concurrent history with a well-formed value relation and without pop-empty 
   operations. *)
locale wellformed_history =
  (* The precedence order *)
  fixes pre :: "'a rel"
  (* The value relation *)
  fixes val :: "'a rel"
  (* The set of operations is countable, f is a witness for this *)
  fixes f :: "'a \<Rightarrow> nat"
  (* The precedence order is a partial order *)
  assumes pre_transitive : "trans pre"
  assumes pre_antisymmetric : "antisym pre"
  assumes pre_irreflexive : "irrefl pre"
  (* The set of operations is countable *)
  assumes operations_countable : "inj f"
  (* The set of operations is finite *)
  assumes operations_finite : "finite {a . isPush val a \<or> isPop val a}"
  (* The value relation is well-formed: *)
  (* Pop operations do not remove more than one element *)
  assumes def_push_once : "push_once val"
  (* No element is popped more than once *)
  assumes def_pop_once : "pop_once val"
  (* An operation either pushes or pops an element, not both *)
  assumes def_value : "val_order val"
  (* The history consists only of push and pop operations*)
  assumes def_push_or_pop : "(a, b) \<in> pre \<Longrightarrow> (\<exists> c. (a, c) \<in> val \<or> (c, a) \<in> val) \<and> (\<exists> c . (b, c) \<in> val \<or> (c, b) \<in> val)"
  (* A pop operation cannot precede its matching push operation *)
  assumes def_val_then_not_pr: "val_then_not_pr pre val" 

lemma wellformed_historyI : "\<And> f::'a\<Rightarrow>nat. trans pre \<Longrightarrow> antisym pre \<Longrightarrow> irrefl pre \<Longrightarrow> inj f \<Longrightarrow> finite {a . isPush val a \<or> isPop val a} \<Longrightarrow>
  push_once val \<Longrightarrow> pop_once val \<Longrightarrow> val_order val \<Longrightarrow> 
  (\<forall> a b .(a, b) \<in> pre \<longrightarrow> (\<exists> c. (a, c) \<in> val \<or> (c, a) \<in> val) \<and> (\<exists> c . (b, c) \<in> val \<or> (c, b) \<in> val)) \<Longrightarrow>
  val_then_not_pr pre val \<Longrightarrow> wellformed_history pre val f"
apply (unfold wellformed_history_def)
by metis

lemma wellformed_historyD : "wellformed_history pre val f \<Longrightarrow> trans pre \<and> antisym pre \<and> irrefl pre \<and> inj f \<and> finite {a . isPush val a \<or> isPop val a} \<and>
  push_once val \<and> pop_once val \<and> val_order val \<and> 
  (\<forall> a b .(a, b) \<in> pre \<longrightarrow> (\<exists> c. (a, c) \<in> val \<or> (c, a) \<in> val) \<and> (\<exists> c . (b, c) \<in> val \<or> (c, b) \<in> val)) \<and>
  val_then_not_pr pre val " 

apply (unfold wellformed_history_def)
by metis

context wellformed_history
begin

definition is_push :: "'a \<Rightarrow> bool"
where
  "is_push a \<longleftrightarrow> (\<exists>b.(a, b)\<in>val)"

definition is_pop :: "'a \<Rightarrow> bool"
where
  "is_pop a \<longleftrightarrow> (\<exists>b.(b, a)\<in>val)"

lemma is_popI : "(ia, ra) \<in> val \<Longrightarrow> is_pop ra"
by (metis is_pop_def)

lemma is_popD : "is_pop ra \<Longrightarrow> \<exists> ia . (ia, ra) \<in> val"
by (metis is_pop_def)

lemma is_pushI : "(ia, ra) \<in> val \<Longrightarrow> is_push ia"
by (metis is_push_def)

lemma is_pushD : "is_push ia \<Longrightarrow> \<exists> ra . (ia, ra) \<in> val"
by (metis is_push_def)

lemma lem_push_or_pop1: "(a, b) \<in> pre \<Longrightarrow> is_push a \<or> is_pop a"
by (metis def_push_or_pop is_popI is_pushI)

lemma lem_push_or_pop2: "(a, b) \<in> pre \<Longrightarrow> is_push b \<or> is_pop b"
by (metis def_push_or_pop is_popI is_pushI)

lemma lem_push_then_pop_exists: "is_push a \<Longrightarrow> \<exists>b.(a, b)\<in>val \<and> is_pop b"
by (metis is_pop_def is_push_def)

lemma lem_pop_then_push_exists: "is_pop a \<Longrightarrow> \<exists>b.(b, a)\<in>val \<and> is_push b"
by (metis is_pop_def is_push_def)

lemma lem_val_then_push_and_pop: "(a, b)\<in>val \<Longrightarrow> (is_push a \<and> is_pop b)"
by (metis is_pop_def is_push_def)

end

lemma val_preserves_ef1 : "wellformed_history pre val f \<Longrightarrow> (a, b) \<in> val \<Longrightarrow> a \<in> elimination_free pre val \<Longrightarrow> b \<in> elimination_free pre val"
by (metis elimination_freeD elimination_freeI2 pop_once_def sameValD val_order_def val_then_not_pr_def wellformed_history.def_pop_once wellformed_history.def_val_then_not_pr wellformed_history.def_value)

lemma val_preserves_ef2 : "wellformed_history pre val f \<Longrightarrow> (a, b) \<in> val \<Longrightarrow> b \<in> elimination_free pre val \<Longrightarrow> a \<in> elimination_free pre val"
by (metis elimination_freeD elimination_freeI1 elimination_freeI2 push_once_def sameValD sameValI1 val_order_def wellformed_history.def_push_once wellformed_history.def_value)

lemma alternating_then_ef_alternating : "wellformed_history pre val f \<Longrightarrow> alternating val ir \<Longrightarrow> alternating (elimination_free_val pre val) (elimination_free_ir pre val ir)"
apply (rule alternatingI)
apply (rule allI)+
apply (rename_tac a b)
apply (rule impI)
apply (drule elimination_free_irD)
apply (erule conjE)+
apply (drule alternatingD)
apply (erule conjE)
apply (erule_tac x="a" in allE)
apply (erule_tac x="a" in allE)
apply (erule_tac x="b" in allE)
apply (erule_tac x="b" in allE)
apply (simp)
apply (erule disjE)
apply (erule conjE)
apply (rule disjI1)
apply (rule conjI)
apply (frule isPushD)
apply (erule exE)
apply (rename_tac ra)
apply (rule_tac b="ra" in isPushI)
apply (metis elimination_free_valI val_preserves_ef1)
apply (metis elimination_free_valI isPop_def val_preserves_ef2)
apply (rule disjI2)
apply (rule conjI)
apply (metis elimination_free_valI isPop_def val_preserves_ef2)
apply (metis elimination_free_valI isPush_def val_preserves_ef1)
apply (rule allI)+
apply (rename_tac a b)
apply (rule impI)+
by (metis (hide_lams, no_types) alternatingD elimination_free_irI elimination_free_valD elimination_free_val_def isPop_def isPush_def)

(* A sequential set is a wellformed history where the precedence relation is total on the set
   of push and pop operations. *)
locale sequential_set = wellformed_history +
  assumes total_on_push : "is_push a \<Longrightarrow> is_push b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre"
  assumes total_on_pop : "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre"
  assumes total_on_push_pop : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre"

lemma sequential_setD : "sequential_set pre val f \<Longrightarrow> wellformed_history pre val f \<and>
  (\<forall> a b . isPush val a \<longrightarrow> isPush val b \<longrightarrow> a \<noteq> b \<longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre) \<and>
  (\<forall> a b . isPop val a \<longrightarrow> isPop val b \<longrightarrow> a \<noteq> b \<longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre)  \<and>
  (\<forall> a b . isPush val a \<longrightarrow> isPop val b \<longrightarrow>(a, b) \<in> pre \<or> (b, a) \<in> pre)"
apply (unfold sequential_set_def)
apply (erule conjE)
apply (rule conjI)
apply (assumption)
apply (unfold sequential_set_axioms_def)
apply (erule conjE)+
apply (rule conjI)
apply (metis isPushD wellformed_history.is_pushI)
by (metis isPopD isPushD wellformed_history.is_popI wellformed_history.is_pushI)

lemma sequential_setI : "wellformed_history pre val f \<Longrightarrow>
  (\<forall> a b . isPush val a \<longrightarrow> isPush val b \<longrightarrow> a \<noteq> b \<longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre) \<Longrightarrow>
  (\<forall> a b . isPop val a \<longrightarrow> isPop val b \<longrightarrow> a \<noteq> b \<longrightarrow> (a, b) \<in> pre \<or> (b, a) \<in> pre)  \<Longrightarrow>
  (\<forall> a b . isPush val a \<longrightarrow> isPop val b \<longrightarrow>(a, b) \<in> pre \<or> (b, a) \<in> pre) \<Longrightarrow> sequential_set pre val f"
apply (unfold sequential_set_def)
apply (rule conjI)
apply (assumption)
apply (unfold sequential_set_axioms_def)
apply (rule conjI)+
apply (metis isPushI wellformed_history.is_pushD)
apply (rule conjI)
apply (metis isPopI wellformed_history.is_popD)
by (metis isPopI isPushI wellformed_history.is_popD wellformed_history.is_pushD)

(* Predicate indicating elements are removed in LIFO order in a sequential history. *)
definition sequential_lifo :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where
  "sequential_lifo pre val \<longleftrightarrow> (\<forall> ia ib ra rb . (ia, ra) \<in> val \<longrightarrow> (ib, rb) \<in> val \<longrightarrow> (ia, ib) \<in> pre \<longrightarrow> (ib, ra) \<in> pre \<longrightarrow> (rb, ra) \<in> pre)"

lemma ef_conc_history : "conc_history pre \<Longrightarrow> conc_history (elimination_free_pre pre val)"
apply (drule conc_historyD)
apply (erule conjE)+
apply (rule conc_historyI)
apply (metis elimination_free_preD elimination_free_preI trans_def)
apply (metis antisym_def elimination_free_preD)
apply (metis elimination_free_preD irrefl_def)
by (smt abcd_def elimination_free_preD elimination_free_preI)

lemma ef_finite: "finite {a. isPush val a \<or> isPop val a} \<Longrightarrow>
finite {a. isPush (elimination_free_val pre val) a \<or>
             isPop (elimination_free_val pre val) a}"
by (metis (lifting, no_types) Collect_mono elimination_free_valD isPop_def isPush_def rev_finite_subset)

lemma val_then_ef_val : "a \<in> elimination_free pre val \<Longrightarrow> (a, b) \<in> val \<Longrightarrow>val_order val \<Longrightarrow> pop_once val \<Longrightarrow>
 (a, b) \<in> elimination_free_val pre val"
apply (drule elimination_freeD)
apply (rule elimination_free_valI)
apply (metis elimination_freeI1 elimination_freeI2 sameValD sameValI1 sameValI2)
apply (metis elimination_freeI1 elimination_freeI2 pop_once_def sameValD sameValI2 val_order_def)
by metis

lemma val_then_not_pr_mono : "val_then_not_pr prt val \<Longrightarrow> pre \<subseteq> prt \<Longrightarrow> val_then_not_pr pre val"
by (metis set_rev_mp val_then_not_pr_def)


definition all_popped_or_empty :: "'a rel \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> bool"
where
  "all_popped_or_empty val emp \<longleftrightarrow> (\<forall>a. emp a \<or> (\<exists> b . (a, b) \<in> val \<or> (b, a) \<in> val))"

definition empty_then_not_val :: "'a rel \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> bool"
where 
  "empty_then_not_val val emp \<longleftrightarrow> (\<forall>a . emp a \<longrightarrow> \<not>(\<exists> b . (a, b) \<in> val) \<and> \<not>(\<exists> b . (b, a) \<in> val))"

(* Definition of a concurrent history with a well-formed value relation and pop-empty predicate. *)
locale wellformed_empty_history =
  (* The precedence order of the history *)
  fixes pre :: "'a rel"
  (* The value relation *)
  fixes val :: "'a rel"
  (* The predicate indicating pop-empty operations *)
  fixes emp :: "'a \<Rightarrow> bool"
  (* The set of operations is countable. The function f is the witness for that. *)
  fixes f :: "'a \<Rightarrow> nat"
  (* The precedence order is a partial order. *)
  assumes pre_transitive : "trans pre"
  assumes pre_antisymmetric : "antisym pre"
  assumes pre_irreflexive : "irrefl pre"
  assumes operations_countable : "inj f"
  (* We only deal with finite histories *)
  assumes operations_finite : "finite {a . isPush val a \<or> isPop val a}"
  (* The set conditions *)
  (* No pop operation removes more than one element *)
  assumes def_push_once : "push_once val"
  (* No element is popped more than once *)
  assumes def_pop_once : "pop_once val"
  (* A push operation cannot be a pop operation, and vice versa*)
  assumes def_val_order : "val_order val"
  (* We only consider histories where all pushed elements also get popped*)
  assumes def_all_popped_or_empty : "all_popped_or_empty val emp"
  (* Consistency condition for the predicate emp *)
  assumes def_empty_then_not_val : "empty_then_not_val val emp"
  (* A pop cannot precede its matching push operation*)
  assumes def_val_then_not_pr : "val_then_not_pr pre val" 

lemma wellformed_empty_historyD : " wellformed_empty_history pre val emp f \<Longrightarrow>
trans pre
\<and> antisym pre
\<and> irrefl pre
\<and> inj f
\<and> finite {a . isPush val a \<or> isPop val a}
\<and>  push_once val
\<and> pop_once val
\<and> val_order val
\<and> all_popped_or_empty val emp
\<and> empty_then_not_val val emp
\<and> val_then_not_pr pre val"
apply (unfold wellformed_empty_history_def)
by metis

lemma wellformed_empty_historyI : 
  "\<And> f::'a\<Rightarrow>nat. 
  trans pre \<Longrightarrow>
  antisym pre \<Longrightarrow>
  irrefl pre \<Longrightarrow>
  inj f \<Longrightarrow>
   finite {a . isPush val a \<or> isPop val a} \<Longrightarrow>
   push_once val \<Longrightarrow>
   pop_once val \<Longrightarrow>
   val_order val \<Longrightarrow>
   all_popped_or_empty val emp \<Longrightarrow>
   empty_then_not_val val emp \<Longrightarrow>
   val_then_not_pr pre val \<Longrightarrow> wellformed_empty_history pre val emp f"
apply (unfold wellformed_empty_history_def)
by metis

context wellformed_empty_history
begin

definition is_push :: "'a \<Rightarrow> bool"
where
  "is_push a \<longleftrightarrow> (\<exists>b.(a, b)\<in>val)"

definition is_pop :: "'a \<Rightarrow> bool"
where
  "is_pop a \<longleftrightarrow> (\<exists>b.(b, a)\<in>val)"

lemma lem_empty_or_push_or_pop : "emp a \<or> is_push a \<or> is_pop a"
by (metis all_popped_or_empty_def def_all_popped_or_empty is_pop_def is_push_def)

lemma lem_push_then_pop_exists: "is_push a \<Longrightarrow> \<exists>b.(a, b)\<in>val \<and> is_pop b"
by (metis is_pop_def is_push_def)

lemma lem_pop_then_push_exists: "is_pop a \<Longrightarrow> \<exists>b.(b, a)\<in>val \<and> is_push b"
by (metis is_pop_def is_push_def)

lemma valD: "(a, b) \<in> val \<Longrightarrow> (is_push a \<and> is_pop b)"
by (metis is_pop_def is_push_def)

end

(* Definition of a sequential history with a well-formed value relation and pop-empty predicate. *)
locale sequential_empty_history = wellformed_empty_history +
  assumes def_total : "total pre"

lemma sequential_empty_historyD : "sequential_empty_history pre val emp f \<Longrightarrow>
wellformed_empty_history pre val emp f
\<and> total pre"
apply (unfold sequential_empty_history_def)
apply (erule conjE)
apply (rule conjI)
apply (assumption)
apply (unfold sequential_empty_history_axioms_def)
by (assumption)

lemma sequential_empty_historyI : "wellformed_empty_history pre val emp f \<Longrightarrow> total pre \<Longrightarrow> 
      sequential_empty_history pre val emp f"
apply (unfold sequential_empty_history_def)
apply (rule conjI)
apply (assumption)
apply (unfold sequential_empty_history_axioms_def)
by (assumption)

(* A predicate indicating correct linearization of pop-empty operations in a sequential history. *)
definition correct_empty :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> bool"
where
  "correct_empty pre val emp \<longleftrightarrow> (\<forall> a . emp a \<longrightarrow> (\<forall> ib rb . (ib, rb) \<in> val \<longrightarrow> ((ib, a) \<in> pre \<and> (rb, a) \<in> pre) \<or> ((a, ib) \<in> pre \<and> (a, rb) \<in> pre)))"

definition non_empty_pre :: "'a rel \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> 'a rel"
where
   "non_empty_pre pre emp = {(a, b) . (a, b) \<in> pre \<and> \<not> emp a \<and> \<not> emp b}"

lemma non_empty_preI : "(a, b) \<in> pre \<Longrightarrow> \<not> emp a \<Longrightarrow> \<not> emp b \<Longrightarrow> (a, b) \<in> non_empty_pre pre emp"
apply (unfold non_empty_pre_def)
by (simp add: set_eq_iff)

lemma non_empty_preD : "(a, b) \<in> non_empty_pre pre emp \<Longrightarrow> \<not> emp a \<and> \<not> emp b \<and> (a, b) \<in> pre"
apply (unfold non_empty_pre_def)
by (simp add: set_eq_iff)

lemma non_empty_pre_then_pre : "(a, b) \<in> non_empty_pre pre emp \<Longrightarrow> (a, b) \<in> pre \<and> \<not> emp a \<and> \<not> emp b"
by (metis (lifting) mem_Collect_eq non_empty_pre_def split_conv)

lemma pre_conc_history_then_non_empty_pre_conc_history : "conc_history pre \<Longrightarrow> conc_history (non_empty_pre pre emp)"
apply (rule conc_historyI)
apply (drule conc_historyD)
apply (erule conjE)+
apply (rule transI)
apply (drule non_empty_pre_then_pre)+
apply (erule conjE)+
apply (metis non_empty_preI transE)
apply (metis antisym_subset conc_historyD non_empty_pre_then_pre subrelI)
apply (metis conc_historyD irrefl_def non_empty_pre_then_pre)
by (smt abcdD abcdI conc_history.abcd non_empty_preI non_empty_pre_then_pre)

lemma ef_all_popped : "\<forall>a b. (a, b) \<in> pre \<longrightarrow>
             (emp a \<or> (\<exists>c. (a, c) \<in> val \<or>
                  (c, a) \<in> val)) \<and>
             (emp b \<or> (\<exists>c. (b, c) \<in> val \<or>
                  (c, b) \<in> val)) \<Longrightarrow>
              val_order val \<Longrightarrow> pop_once val \<Longrightarrow> push_once val \<Longrightarrow>
              (a, b) \<in> elimination_free_pre (non_empty_pre pre emp) val \<Longrightarrow>
             (\<exists>c. (a, c) \<in> elimination_free_val (non_empty_pre pre emp) val \<or>
            (c, a) \<in> elimination_free_val (non_empty_pre pre emp) val) \<and>
       (\<exists>c. (b, c) \<in> elimination_free_val (non_empty_pre pre emp) val \<or>
            (c, b) \<in> elimination_free_val (non_empty_pre pre emp) val)"
apply (drule elimination_free_preD)
apply (erule conjE)
apply (erule_tac x="a" in allE)
apply (erule_tac x="b" in allE)
apply (erule conjE)
apply (erule impE)
apply (metis non_empty_preD)
apply (erule conjE)
apply (rule conjI)
apply (erule disjE)
apply (metis non_empty_preD)
apply (erule exE)+
apply (rule_tac x="c" in exI)
apply (erule disjE)
apply (rule disjI1)
apply (rule elimination_free_valI)
apply (assumption)
apply (metis non_empty_preD)
apply (metis non_empty_preD)
apply (metis elimination_freeD elimination_freeI1 elimination_freeI2 push_once_def sameValD sameValI1 val_order_def val_then_ef_val)
by (metis elimination_freeD elimination_freeI1 elimination_freeI2 non_empty_pre_then_pre push_once_def sameValD sameValI1 val_then_ef_val)

lemma ef_empty_mono : "pre \<subseteq> prt \<Longrightarrow> (elimination_free_pre (non_empty_pre pre emp) val) \<subseteq> prt"
apply (rule subrelI)
apply (drule elimination_free_preD)
apply (erule conjE)+
by (metis non_empty_preD set_rev_mp)

lemma ef_val_then_val : "(a, b) \<in> (elimination_free_val (non_empty_pre pre emp) val) \<Longrightarrow> (a, b) \<in> val"
by (metis elimination_free_valD)

lemma ef_push_then_empty_push : "isPush (elimination_free_val pre val) a \<Longrightarrow> empty_then_not_val val emp \<Longrightarrow>
       isPush (elimination_free_val (non_empty_pre pre emp) val) a"
apply (drule isPushD)
apply (erule exE)
apply (rule_tac b="b" in isPushI)
apply (drule elimination_free_valD)
apply (erule conjE)+
apply (rule elimination_free_valI)
apply (drule elimination_freeD)
apply (erule exE)
apply (erule conjE)
apply (erule disjE)
apply (rule_tac b="ba" in elimination_freeI1)
apply (assumption)
apply (metis empty_then_not_val_def non_empty_preI sameValD)
apply (metis (hide_lams, mono_tags) elimination_freeI2 empty_then_not_val_def non_empty_preI sameValD sameValI1 sameValI2)
apply (drule_tac a="b" in elimination_freeD)
apply (erule exE)
apply (erule conjE)
apply (erule disjE)
apply (rule_tac b="ba" in elimination_freeI1)
apply (assumption)
apply (metis empty_then_not_val_def non_empty_preI sameValD)
apply (metis (hide_lams, mono_tags) elimination_freeI2 empty_then_not_val_def non_empty_preI sameValD sameValI1 sameValI2)
by assumption

lemma empty_push_then_ef_push : "isPush (elimination_free_val (non_empty_pre pre emp) val) a \<Longrightarrow>
  isPush (elimination_free_val pre val) a"
apply (drule isPushD)
apply (erule exE)
apply (rule_tac b="b" in isPushI)
apply (drule elimination_free_valD)
apply (erule conjE)+
apply (rule elimination_free_valI)
apply (metis elimination_freeD elimination_freeI1 elimination_freeI2 non_empty_pre_then_pre sameVal_commutative)
apply (metis elimination_freeD elimination_freeI1 elimination_freeI2 non_empty_pre_then_pre sameVal_commutative)
by metis

lemma ef_pop_then_empty_pop : "isPop (elimination_free_val pre val) a \<Longrightarrow> empty_then_not_val val emp \<Longrightarrow> val_order val \<Longrightarrow> pop_once val \<Longrightarrow> push_once val \<Longrightarrow>
       isPop (elimination_free_val (non_empty_pre pre emp) val) a"
apply (drule isPopD)
apply (erule exE)
apply (rule_tac a="b" in isPopI)
apply (drule elimination_free_valD)
apply (erule conjE)+
apply (rule elimination_free_valI)
apply (drule elimination_freeD)
apply (erule exE)
apply (erule conjE)
apply (erule disjE)
apply (rule_tac b="ba" in elimination_freeI1)
apply (assumption)
apply (metis empty_then_not_val_def non_empty_preI sameValD)
apply (metis (hide_lams, mono_tags) elimination_freeI2 empty_then_not_val_def non_empty_preI sameValD sameValI1 sameValI2)
apply (drule_tac a="b" in elimination_freeD)
apply (erule exE)
apply (erule conjE)
apply (erule disjE)
apply (rule_tac a="b" in elimination_freeI2)
apply (metis sameValI1)
apply (metis empty_then_not_val_def non_empty_preI pop_once_def sameValD val_order_def)
apply (metis (full_types) elimination_freeI1 empty_then_not_val_def non_empty_preI pop_once_def sameValD sameVal_commutative val_order_def)
by assumption

lemma empty_pop_then_ef_pop : "isPop (elimination_free_val (non_empty_pre pre emp) val) a \<Longrightarrow>
  isPop (elimination_free_val pre val) a"
apply (drule isPopD)
apply (erule exE)
apply (rule_tac a="b" in isPopI)
apply (drule elimination_free_valD)
apply (erule conjE)+
apply (rule elimination_free_valI)
apply (metis elimination_freeD elimination_freeI1 elimination_freeI2 non_empty_pre_then_pre sameVal_commutative)
apply (metis elimination_freeD elimination_freeI1 elimination_freeI2 non_empty_pre_then_pre sameVal_commutative)
by metis

lemma ef_ins_then_ins : "(ia, ib)
       \<in> ins_order (elimination_free_pre (non_empty_pre pre emp) val)
           (elimination_free_val (non_empty_pre pre emp) val) ir \<Longrightarrow> (ia, ib) \<in> ins_order pre val ir"
apply (drule ins_orderD)
apply (erule conjE)+
apply (erule disjE)
apply (metis ef_val_then_val elimination_free_preD ins_orderI1 isPushD isPushI non_empty_pre_then_pre)
by (metis ef_val_then_val elimination_free_preD ins_orderI2 isPushD isPushI non_empty_pre_then_pre)

lemma ef_rem_then_rem : "(ra, rb)
       \<in> rem_order (elimination_free_pre (non_empty_pre pre emp) val)
           (elimination_free_val (non_empty_pre pre emp) val) ir \<Longrightarrow> (ra, rb) \<in> rem_order pre val ir"
apply (drule rem_orderD)
apply (erule conjE)+
apply (erule disjE)
apply (metis ef_val_then_val elimination_free_preD isPopD isPopI non_empty_pre_then_pre rem_orderI1)
by (metis ef_val_then_val isPopD isPopI rem_orderI2)

(* A sequential stack is a sequential set where elements are removed in LIFO order *)
locale sequential_stack = sequential_set + 
  assumes def_sequential_lifo : "sequential_lifo pre val"

lemma sequential_stackD : "sequential_stack pre val f \<Longrightarrow> sequential_set pre val f \<and> sequential_lifo pre val"
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
by metis

lemma sequential_stackI : "sequential_set pre val f \<Longrightarrow> sequential_lifo pre val \<Longrightarrow> sequential_stack pre val f"
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
by metis

(* Definition of a sequential set with pop-empty operations *)
locale sequential_empty_set = sequential_empty_history +
    assumes def_correct_empty : "correct_empty pre val emp"

lemma sequential_empty_setD : "sequential_empty_set pre val emp f \<Longrightarrow>
sequential_empty_history pre val emp f \<and>
correct_empty pre val emp"
apply (unfold sequential_empty_set_def)
apply (unfold sequential_empty_set_axioms_def)
by metis

lemma sequential_empty_setI : "sequential_empty_history pre val emp f \<Longrightarrow> correct_empty pre val emp \<Longrightarrow> 
      sequential_empty_set pre val emp f"
apply (unfold sequential_empty_set_def)
apply (unfold sequential_empty_set_axioms_def)
by metis

(* A sequential stack with pop-empty is a sequential set where elements are removed in LIFO order.*)
locale sequential_empty_stack = sequential_empty_history + 
  assumes def_sequential_lifo : "sequential_lifo pre val"
  assumes def_correct_empty : "correct_empty pre val emp"

lemma sequential_empty_stackD : "sequential_empty_stack pre val emp f \<Longrightarrow>
sequential_empty_history pre val emp f \<and>
sequential_lifo pre val \<and>
correct_empty pre val emp"
apply (unfold sequential_empty_stack_def)
apply (unfold sequential_empty_stack_axioms_def)
by metis

lemma sequential_empty_stackI : "sequential_empty_history pre val emp f \<Longrightarrow>
sequential_lifo pre val \<Longrightarrow>
correct_empty pre val emp \<Longrightarrow>
sequential_empty_stack pre val emp f"
apply (unfold sequential_empty_stack_def)
apply (unfold sequential_empty_stack_axioms_def)
by metis

(* A history <A, pre, val> is linearizable with respect to SET if it is a valid concurrent history 
   (conc_history pre) and there exists a linearization order prt such that <A, prt, val> is in SET
   (sequential_empty_set prt val emp f). The function f indicates that the set of operations is
   countable, the predicate emp indicates pop-empty operations. *)
definition linearizable_set :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> bool"
where
  "linearizable_set pre val emp \<longleftrightarrow> conc_history pre \<and> (\<exists> prt f . pre \<subseteq> prt \<and> sequential_empty_set prt val emp f)"

lemma linearizable_setD : "linearizable_set pre val emp \<Longrightarrow> conc_history pre \<and>(\<exists> prt f . pre \<subseteq> prt \<and> sequential_empty_set prt val emp f)"
by (smt linearizable_set_def)

lemma linearizable_setI : "conc_history pre \<Longrightarrow>  pre \<subseteq> prt \<Longrightarrow> sequential_empty_set prt val emp f \<Longrightarrow> linearizable_set pre val emp"
by (metis linearizable_set_def)

definition linearizable_non_empty_stack :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where
  "linearizable_non_empty_stack pre val \<longleftrightarrow> conc_history pre \<and> (\<exists> prt f . pre \<subseteq> prt \<and> sequential_stack prt val f)"

lemma linearizable_non_empty_stackI : "conc_history pre \<Longrightarrow> pre \<subseteq> prt \<Longrightarrow> sequential_stack prt val f \<Longrightarrow> linearizable_non_empty_stack pre val"
by (metis linearizable_non_empty_stack_def)

lemma linearizable_non_empty_stackD : "linearizable_non_empty_stack pre val \<Longrightarrow> conc_history pre \<and> (\<exists> prt f . pre \<subseteq> prt \<and> sequential_stack prt val f)"
by (smt linearizable_non_empty_stack_def)

definition linearizable_stack :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> bool"
where
  "linearizable_stack pre val emp \<longleftrightarrow> conc_history pre \<and> (\<exists> prt f. pre \<subseteq> prt \<and> sequential_empty_stack prt val emp f)"

lemma linearizable_stackD : "linearizable_stack pre val emp \<Longrightarrow> conc_history pre \<and> (\<exists> prt f. pre \<subseteq> prt \<and> sequential_empty_stack prt val emp f)"
by (smt linearizable_stack_def)

lemma linearizable_stackI : "conc_history pre \<Longrightarrow> pre \<subseteq> prt \<Longrightarrow> sequential_empty_stack prt val emp f \<Longrightarrow> linearizable_stack pre val emp"
by (metis linearizable_stack_def)

lemma linearizable_set_then_wellformed_empty : "linearizable_set pre val emp \<Longrightarrow> \<exists> f . wellformed_empty_history pre val emp f"
apply (drule linearizable_setD)
apply (erule conjE)
apply (erule exE)+
apply (erule conjE)
apply (drule sequential_empty_setD)
apply (erule conjE)
apply (drule sequential_empty_historyD)
apply (erule conjE)
apply (rule_tac x="f" in exI)
apply (rule wellformed_empty_historyI)
apply (metis conc_historyD)
apply (metis conc_history.pre_antisymmetric)
apply (metis conc_history.pre_irreflexive)
apply (metis wellformed_empty_history.operations_countable)
apply (metis wellformed_empty_historyD)
apply (metis wellformed_empty_history.def_push_once)
apply (metis wellformed_empty_history.def_pop_once)
apply (metis wellformed_empty_history.def_val_order)
apply (metis wellformed_empty_history.def_all_popped_or_empty)
apply (metis wellformed_empty_history.def_empty_then_not_val)
by (metis val_then_not_pr_mono wellformed_empty_history.def_val_then_not_pr)

lemma wellformed_empty_then_wellformed : "wellformed_empty_history pre val emp f \<Longrightarrow> wellformed_history (non_empty_pre pre emp) val f"
apply (drule wellformed_empty_historyD)
apply (erule conjE)+
apply (rule wellformed_historyI)
apply (metis (lifting, no_types) non_empty_preI non_empty_pre_then_pre trans_def)
apply (metis (lifting, no_types) antisymI irrefl_def non_empty_pre_then_pre trans_def)
apply (metis irrefl_def non_empty_preD)
apply metis
apply metis
apply metis
apply metis
apply metis
apply (metis (hide_lams, no_types) all_popped_or_empty_def non_empty_pre_then_pre)
by (metis non_empty_pre_then_pre val_then_not_pr_def)

lemma linearizable_set_then_wellformed : "linearizable_set pre val emp \<Longrightarrow> \<exists> f .wellformed_history (non_empty_pre pre emp) val f"
by (metis linearizable_set_then_wellformed_empty wellformed_empty_then_wellformed)

lemma wellformed_then_ef_wellformed : "wellformed_history pre val f \<Longrightarrow> wellformed_history (elimination_free_pre pre val) (elimination_free_val pre val) f"
apply (drule wellformed_historyD)
apply (erule conjE)+
apply (rule wellformed_historyI)
apply (metis (no_types, hide_lams) elimination_free_preD elimination_free_preI transE transI)
apply (metis (poly_guards_query) antisym_def elimination_free_preD)
apply (metis elimination_free_preD irrefl_def)
apply metis
apply (metis ef_finite)
apply (metis elimination_free_valD push_once_def)
apply (metis (full_types) elimination_free_valD pop_once_def)
apply (metis (full_types) elimination_free_valD val_order_def)
apply (rule allI)+
apply (rule impI)
apply (erule_tac x="a" in allE)
apply (erule_tac x="b" in allE)
apply (frule elimination_free_preD)
apply (erule conjE)+
apply (erule impE)
apply (assumption)
apply (erule conjE)
apply (erule exE)+
apply (erule disjE)+
apply (rule conjI)
apply (rule_tac x="c" in exI)
apply (metis val_then_ef_val)
apply (rule_tac x="ca" in exI)
apply (metis val_then_ef_val)
apply (rule conjI)
apply (rule_tac x="c" in exI)
apply (metis val_then_ef_val)
apply (rule_tac x="ca" in exI)
apply (metis (no_types, hide_lams) elimination_freeD elimination_freeI1 push_once_def sameValD sameValI1 val_order_def val_then_ef_val val_then_not_pr_def)
apply (rule conjI)
apply (rule_tac x="c" in exI)
apply (metis (no_types, hide_lams) elimination_freeD elimination_freeI1 push_once_def sameValD sameValI1 val_order_def val_then_ef_val val_then_not_pr_def)
apply (rule_tac x="ca" in exI)
apply (erule disjE)
apply (metis val_then_ef_val)
apply (metis (no_types, hide_lams) elimination_freeD elimination_freeI1 push_once_def sameValD sameValI1 val_order_def val_then_ef_val val_then_not_pr_def)
by (metis elimination_free_preD elimination_free_valD val_then_not_pr_def)

lemma ef_ir_in_ir : "elimination_free_ir (non_empty_pre pre emp) val ir \<subseteq> ir"
by (metis elimination_free_irD subrelI)

lemma help_me : "pre Un (elimination_free_ir (non_empty_pre pre emp) val ir) \<subseteq> pre Un ir"
by (metis ef_ir_in_ir order_refl sup.mono)

lemma ef_val_then_pre : "wellformed_history (non_empty_pre pre emp) val f \<Longrightarrow> (ia, ra) \<in> elimination_free_val (non_empty_pre pre emp) val \<Longrightarrow> (ia, ra) \<in> pre"
apply (drule elimination_free_valD)
apply (erule conjE)+
apply (drule elimination_freeD)+
apply (erule exE)+
apply (erule conjE)+
apply (erule disjE)+
apply (metis (erased, hide_lams) sameValD val_order_def val_then_not_pr_def wellformed_history.def_val_then_not_pr wellformed_history.def_value)
apply (metis (erased, hide_lams) non_empty_pre_then_pre push_once_def sameValD val_order_def wellformed_history.def_push_once wellformed_history.def_value)
by (metis (erased, hide_lams) sameValD val_order_def val_then_not_pr_def wellformed_history.def_val_then_not_pr wellformed_history.def_value)

lemma elim_ef_rem_then_rem: "(ra, rb) \<in> rem_order (elimination_free_pre pre val) (elimination_free_val pre val) (elimination_free_ir pre val ir) \<Longrightarrow>
       wellformed_history pre val f \<Longrightarrow>
       (ra, rb) \<in> rem_order pre val ir"
apply (drule rem_orderD)
apply (erule conjE)+
apply (erule disjE)
apply (rule rem_orderI1)
apply (metis elimination_free_valD isPop_def)
apply (metis elimination_free_valD isPop_def)
apply (metis elimination_free_preD)
apply (erule exE)
apply (erule conjE)
apply (rule_tac c="c" in rem_orderI2)
apply (metis elimination_free_valD isPop_def)
apply (metis elimination_free_valD isPop_def)
apply (metis elimination_free_irD)
by (metis elimination_free_irD)

lemma empty_rem_then_rem : "(ra, rb) \<in> rem_order (non_empty_pre pre emp) val ir \<Longrightarrow> wellformed_history (non_empty_pre pre emp) val f \<Longrightarrow> (ra, rb) \<in> rem_order pre val ir"
by (metis (erased, hide_lams) non_empty_preD rem_orderD rem_orderI1 rem_orderI2)

end
